from __future__ import annotations

import builtins
import dataclasses
import io
import sys
import typing
import types
from typing import *
from typing_extensions import ParamSpec, Annotated

import pytest

import introspection
from introspection.errors import CannotResolveForwardref
from introspection.typing import ForwardRef, TypeInfo
from introspection.typing.misc import *
from introspection.typing.misc2 import *


T = TypeVar("T")
RecursiveType = List["RecursiveType"]

THIS_MODULE = sys.modules[__name__]


@pytest.mark.parametrize(
    "expected",
    [
        "int",
        "List",
        "io.IOBase",
        "List[str]",
        "Tuple[List[int], bool]",
        "...",
        "Callable[[int], None]",
    ],
)
def test_annotation_to_string_simple(expected):
    annotation = eval(expected)

    assert annotation_to_string(annotation) == expected


if sys.version_info >= (3, 10):

    @pytest.mark.parametrize(
        "expected",
        [
            "int | str",
            "int | None",
            "int | str | None",
        ],
    )
    def test_annotation_to_string_simple_py310(expected: str):
        annotation = eval(expected)

        assert annotation_to_string(annotation) == expected

    @pytest.mark.parametrize(
        "annotation, expected",
        [
            (dataclasses.KW_ONLY, "KW_ONLY"),
        ],
    )
    def test_annotation_to_string_py310(annotation, expected: str):
        assert annotation_to_string(annotation) == expected


@pytest.mark.parametrize(
    "expected",
    [
        "Union[int, str]",
        "Union[int, str, None]",
    ],
)
def test_annotation_to_string_old_style_unions(expected):
    annotation = eval(expected)

    assert annotation_to_string(annotation, new_style_unions=False) == expected


@pytest.mark.parametrize(
    "annotation, expected",
    [
        (None, "None"),
        (type(None), "None"),
        (List["Foo"], "List[Foo]"),  # type: ignore[foo-not-defined]
        (TypeVar("T"), "T"),
        (TypeVar("T", covariant=True), "T"),
        (TypeVar("T", contravariant=True), "T"),
        (ParamSpec("P"), "P"),
        (ParamSpec("P").args, "P.args"),  # type: ignore
        (ParamSpec("P").kwargs, "P.kwargs"),  # type: ignore
    ],
)
def test_annotation_to_string(annotation, expected):
    assert annotation_to_string(annotation) == expected


@pytest.mark.parametrize(
    "annotation, expected",
    [
        (TypeVar("T"), "~T"),
        (TypeVar("T", covariant=True), "+T"),
        (TypeVar("T", contravariant=True), "-T"),
        (ParamSpec("P"), "~P"),
        (ParamSpec("P").args, "~P.args"),  # type: ignore
        (ParamSpec("P").kwargs, "~P.kwargs"),  # type: ignore
    ],
)
def test_annotation_to_string_with_variance_prefixes(annotation, expected):
    assert annotation_to_string(annotation, variance_prefixes=True) == expected


if hasattr(typing, "Literal"):

    @pytest.mark.parametrize(
        "expected",
        [
            "Literal[1, 3, 'foo']",
        ],
    )
    def test_literal_to_string(expected):
        annotation = eval(expected)

        assert annotation_to_string(annotation) == expected


if hasattr(typing, "TypeVarTuple"):

    @pytest.mark.parametrize(
        "annotation, expected",
        [
            (TypeVarTuple("T"), "T"),  # type: ignore
        ],
    )
    def test_typevartuple_to_string(annotation, expected):
        assert annotation_to_string(annotation) == expected


@pytest.mark.parametrize(
    "annotation, expected",
    [
        (int, "int"),
        (List, "typing.List"),
        (List[str], "typing.List[str]"),
        (List[Callable], "typing.List[typing.Callable]"),
    ],
)
def test_annotation_to_string_with_typing(annotation, expected):
    ann = annotation_to_string(annotation, implicit_typing=False)

    assert ann == expected


if hasattr(typing, "Literal"):

    @pytest.mark.parametrize(
        "annotation, expected",
        [
            (Literal[1, 3, "foo"], "typing.Literal[1, 3, 'foo']"),
        ],
    )
    def test_literal_to_string_with_typing(annotation, expected):
        ann = annotation_to_string(annotation, implicit_typing=False)

        assert ann == expected


@pytest.mark.parametrize(
    "annotation, expected",
    [
        (int, int),
        (List, List),
        (List["str"], List[str]),
        (Tuple[Dict["str", "int"]], Tuple[Dict[str, int]]),
        (Tuple["Dict[str, int]"], Tuple[Dict[str, int]]),
        (Callable[["int"], str], Callable[[int], str]),
        ("ellipsis", type(...)),
        ('List["int"]', List[int]),
        ('Literal["int"]', Literal["int"]),  # `Literal` arguments must be left as strings
        ('"int"', int),  # Double stringified
        ("RecursiveType", RecursiveType),
    ],
)
@pytest.mark.parametrize("mode", ["eval", "ast"])
def test_resolve_forward_refs(mode, annotation, expected):
    assert resolve_forward_refs(annotation, globals(), mode=mode) == expected


@pytest.mark.parametrize(
    "annotation, expected",
    [
        ("int if False else float", float),
    ],
)
def test_resolve_forward_refs_eval(annotation, expected):
    assert resolve_forward_refs(annotation, globals(), mode="eval") == expected


# TypeVars can't be compared with ==, so they need a specialized test function
@pytest.mark.parametrize(
    "annotation, expected",
    [
        (TypeVar("T", bound="int"), TypeVar("T", bound=int)),
        (TypeVar("T", bound="int", covariant=True), TypeVar("T", bound=int, covariant=True)),
        (TypeVar("T", "int", "str"), TypeVar("T", int, str)),
    ],
)
def test_resolve_typevar_forward_refs(annotation, expected: TypeVar):
    result = cast(TypeVar, resolve_forward_refs(annotation))

    assert result.__name__ == expected.__name__
    assert result.__bound__ == expected.__bound__
    assert result.__constraints__ == expected.__constraints__
    assert result.__covariant__ == expected.__covariant__
    assert result.__contravariant__ == expected.__contravariant__


if hasattr(typing, "Literal"):

    @pytest.mark.parametrize(
        "annotation, expected",
        [
            (Literal["int", "Tuple"], Literal["int", "Tuple"]),
            # (Literal[ForwardRef('int'), 'Tuple'], Literal[int, 'Tuple']),
        ],
    )
    def test_resolve_forward_refs_literal(annotation, expected):
        assert resolve_forward_refs(annotation) == expected


@pytest.mark.parametrize(
    "annotation",
    [
        "enum.Enum",
        "types.FunctionType",
        "queue.Queue",
    ],
)
@pytest.mark.parametrize("mode", ["eval", "ast", "getattr"])
def test_resolve_forward_refs_with_missing_imports(mode, annotation):
    result = resolve_forward_refs(annotation, mode=mode, treat_name_errors_as_imports=True)
    assert not isinstance(result, str)


memoryview = bytearray


@pytest.mark.parametrize(
    "annotation, module, expected",
    [
        ("memoryview", THIS_MODULE, bytearray),
        ("memoryview", None, builtins.memoryview),
    ],
)
def test_resolve_forward_refs_in_module(annotation, module, expected):
    ann = resolve_forward_refs(annotation, module)
    assert ann == expected


@pytest.mark.parametrize(
    "annotation, module, expected",
    [
        ("io.IOBase", THIS_MODULE, io.IOBase),
    ],
)
def test_resolve_forward_refs_no_eval(annotation, module, expected):
    ann = resolve_forward_refs(annotation, module, mode="getattr")
    assert ann == expected


@pytest.mark.parametrize(
    "annotation, kwargs",
    [
        ("ThisClassDoesntExist", {"mode": "getattr"}),
        ("ThisClassDoesntExist", {"context": "sys"}),
        ("this is a syntax error", {}),
    ],
)
def test_resolve_forward_refs_error(annotation, kwargs):
    with pytest.raises(ValueError):
        resolve_forward_refs(annotation, **kwargs)


@pytest.mark.parametrize(
    "annotation, kwargs, expected",
    [
        ("ThisClassDoesntExist", {"mode": "getattr"}, "ThisClassDoesntExist"),
        ("this is a syntax error", {}, "this is a syntax error"),
        ('List["Foo"]', {}, List["Foo"]),  # type: ignore[Foo-not-defined]
    ],
)
def test_resolve_forward_refs_non_strict(annotation, kwargs, expected):
    ann = resolve_forward_refs(annotation, strict=False, **kwargs)
    assert ann == expected


@pytest.mark.parametrize(
    "annotation, expected",
    [
        ("List[Foo]", List["Foo"]),  # type: ignore[Foo-not-defined]
        ("Callable[[Foo], str]", Callable[["Foo"], str]),  # type: ignore[Foo-not-defined]
    ],
)
def test_partially_resolve_forward_refs(annotation: str, expected):
    ann = resolve_forward_refs(annotation, globals(), mode="ast", strict=False)
    assert ann == expected


def test_resolve_annotated():
    ann = resolve_forward_refs('Annotated[int, "hi"]', globals(), mode="ast", strict=False)
    assert ann == Annotated[int, "hi"]


def no_args():
    pass


def nil__nil(x):
    pass


def args_nil(*args):
    pass


def str_args__float(foo: str, *args) -> float:
    return 3


def t_args__t(foo: T, *args) -> T:
    return foo


@pytest.mark.parametrize(
    "callable_, expected",
    [
        (len, Callable[[Sized], int]),
        (
            range,
            Union[
                Callable[[int, int, int], range],
                Callable[[int, int], range],
                Callable[[int], range],
            ],
        ),
        (nil__nil, Callable[[Any], Any]),
        (no_args, Callable[[], Any]),
        # (args_nil, Callable),
        # (str_args__float, Callable[..., float]),
        # (t_args__t, Union[
        #     Callable[[T], T],
        #     Callable[..., T],
        # ]),
    ],
)
def test_annotation_for_callable(callable_, expected):
    ann = annotation_for_callable(callable_)
    assert ann == expected


if hasattr(dataclasses, "KW_ONLY"):

    def test_kw_only():
        @dataclasses.dataclass
        class Foo:
            foo: int
            _: dataclasses.KW_ONLY
            bar: str

        assert get_type_annotations(Foo)["_"].type is dataclasses.KW_ONLY


if hasattr(dataclasses, "InitVar"):

    def test_initvar():
        @dataclasses.dataclass
        class Foo:
            foo: dataclasses.InitVar[int]

        assert get_type_annotations(Foo)["foo"].type is dataclasses.InitVar


# The parent class has an attribute annotated with a forward reference. That forward reference will
# be copied to the constructor of the child class. Check if it can be resolved.
from . import utils


@dataclasses.dataclass
class Child(utils.Parent):
    bar: str


def test_resolve_forward_ref_from_parent_class():
    signature = introspection.Signature.from_callable(Child)
    parameter = signature.parameters["foo"]
    annotation = resolve_forward_refs(
        parameter.annotation,
        parameter.forward_ref_context,
    )
    assert annotation is utils.SomeTypeDefinedInParentsFile

    # Make sure it also works when we ask for the signature of `__init__` directly
    signature = introspection.Signature.from_callable(Child.__init__)
    parameter = signature.parameters["foo"]
    annotation = resolve_forward_refs(
        parameter.annotation,
        parameter.forward_ref_context,
    )
    assert annotation is utils.SomeTypeDefinedInParentsFile


def test_get_type_annotations():
    class Foo:
        foo: int | str

    annotations = get_type_annotations(Foo)
    assert list(annotations) == ["foo"]
    assert annotations["foo"].type in (typing.Union, types.UnionType)
    assert annotations["foo"].arguments == (int, str)


def test_get_instance_attribute_annotations():
    class Foo:
        a: int
        b: ClassVar[int]
        c: ClassVar[str] = "hi"
        d: int = 5

    annotations = get_instance_attribute_annotations(Foo)
    assert list(annotations) == ["a", "d"]
    assert annotations["a"].type is int
    assert annotations["d"].type is int


if hasattr(dataclasses, "KW_ONLY"):

    def test_get_instance_attribute_annotations_kw_only():
        @dataclasses.dataclass
        class Foo:
            a: int
            _: dataclasses.KW_ONLY
            b: str

        annotations = get_instance_attribute_annotations(Foo)
        assert list(annotations) == ["a", "b"]
        assert annotations["a"].type is int
        assert annotations["b"].type is str


def test_get_instance_attribute_annotations_forward_refs():
    class Foo:
        a: "Undefined"  # type: ignore
        b: "ClassVar[Undefined]"  # type: ignore

    with pytest.raises(CannotResolveForwardref):
        get_instance_attribute_annotations(Foo)

    annotations = get_instance_attribute_annotations(Foo, allow_forwardrefs=True)
    assert list(annotations) == ["a"]
    assert isinstance(annotations["a"], ForwardRef)


def test_get_instance_attribute_annotations_explicit_false():
    class Foo:
        a: int

    annotations = get_instance_attribute_annotations(Foo, allow_forwardrefs=False)
    assert list(annotations) == ["a"]
    assert annotations["a"].type is int


def test_get_instance_attribute_annotations_explicit_true_no_forward_refs():
    class Foo:
        a: int

    annotations = get_instance_attribute_annotations(Foo, allow_forwardrefs=True)
    assert list(annotations) == ["a"]
    assert isinstance(annotations["a"], TypeInfo)
    assert annotations["a"].type is int



from __future__ import annotations

import ast
import builtins
import collections.abc
import dataclasses
import importlib
import types
import typing  # For runtime
import typing as t  # For annotations
import warnings

import typing_extensions

from ._compat import LITERAL_TYPES
from .introspection import (
    is_parameterized_generic,
    get_generic_base_class,
    get_type_arguments,
    get_type_name,
    _get_forward_ref_code,
)
from .i_hate_circular_imports import parameterize
from ..parameter import Parameter
from ..signature_ import Signature
from ..types import Type_, TypeAnnotation, ForwardRefContext
from ..errors import *

__all__ = [
    "resolve_forward_refs",
    "annotation_to_string",
    "annotation_for_callable",
]


KW_ONLY = getattr(dataclasses, "KW_ONLY", object())


@t.overload
def resolve_forward_refs(
    annotation: TypeAnnotation,
    context: t.Optional[ForwardRefContext] = None,
    *,
    mode: t.Literal["eval", "getattr", "ast"] = "eval",
    strict: bool = True,
    max_depth: t.Optional[int] = None,
    extra_globals: t.Mapping[str, object] = {},
    treat_name_errors_as_imports: bool = False,
) -> TypeAnnotation: ...


@t.overload
def resolve_forward_refs(
    annotation: TypeAnnotation,
    module: t.Optional[types.ModuleType] = None,
    eval_: bool = True,
    strict: bool = True,
) -> TypeAnnotation: ...


def resolve_forward_refs(  # type: ignore
    annotation: TypeAnnotation,
    context: t.Optional[ForwardRefContext] = None,
    eval_: t.Optional[bool] = None,
    strict: bool = True,
    *,
    module: t.Optional[types.ModuleType] = None,
    mode: t.Literal["eval", "getattr", "ast"] = "eval",
    max_depth: t.Optional[int] = None,
    extra_globals: t.Mapping[str, object] = {},
    treat_name_errors_as_imports: bool = False,
    _currently_evaluating: t.AbstractSet[t.Tuple[str, int]] = set(),
) -> TypeAnnotation:
    """
    Resolves forward references in a type annotation.

    Examples::

        >>> resolve_forward_refs(List['int'])
        typing.List[int]
        >>> resolve_forward_refs('ellipsis')
        <class 'ellipsis'>

    Using `mode='ast'` makes partial evaluation possible::

        >>> resolve_forward_refs('List[ThisCantBeResolved'], mode='ast', strict=False)
        List['ThisCantBeResolved']

    .. versionchanged:: 1.6
        The ``module`` and ``eval_`` parameters are deprecated in favor of
        ``context`` and ``mode``.

    :param annotation: The annotation in which forward references should be resolved
    :param context: The context in which forward references will be evaluated.
        Can be a class, a function, a module, or a string representing a module
        name. If ``None``, a namespace containing some common modules like
        ``typing`` and ``collections`` will be used.
    :param mode: If ``'eval'``, references may contain arbitrary code that will
        be evaluated with ``eval``. If ``'getattr'``, they must be identifiers
        and will be resolved with ``getattr``.
    :param strict: Whether to raise an exception if a forward reference can't be resolved
    :return: A new annotation with no forward references
    """
    if module is not None:
        context = module
        warnings.warn(
            'The "module" parameter is deprecated in favor of "context"',
            DeprecationWarning,
        )

    if eval_ is not None:
        mode = "eval" if eval_ else "ast"
        warnings.warn(
            'The "eval_" parameter is deprecated in favor of "mode"',
            DeprecationWarning,
        )

    if max_depth is None:
        max_depth = t.cast(int, float("inf"))
    elif max_depth <= 0:
        return annotation

    def recurse(annotation: TypeAnnotation) -> TypeAnnotation:
        return resolve_forward_refs(
            annotation,
            context,
            mode=mode,
            strict=strict,
            max_depth=max_depth - 1,
            extra_globals=extra_globals,
            _currently_evaluating=_currently_evaluating,  # type: ignore
        )

    if isinstance(annotation, t.ForwardRef):
        if context is None:
            context = annotation.__forward_module__

        annotation = _get_forward_ref_code(annotation)

    if isinstance(annotation, str):
        # First, check if this exact same forward reference is already being evaluated - i.e. it is
        # a recursive type.
        key = (annotation, id(context))
        if key in _currently_evaluating:
            return annotation
        _currently_evaluating = _currently_evaluating | {key}

        scope: collections.ChainMap[str, object] = collections.ChainMap()

        if context is None:
            scope.maps.extend(
                vars(module)
                for module in (collections.abc, collections, typing, typing_extensions)  # type: ignore
            )
        elif isinstance(context, types.ModuleType):
            scope.maps.append(vars(context))
        elif isinstance(context, str):
            module = importlib.import_module(context)
            scope.maps.append(vars(module))
        elif isinstance(context, collections.abc.Mapping):
            scope.maps.append(context)  # type: ignore
        else:
            module = importlib.import_module(context.__module__)
            scope.maps.append(vars(module))

        scope.maps.append(vars(builtins))  # type: ignore
        scope.maps.append(extra_globals)  # type: ignore

        if treat_name_errors_as_imports:
            from ._utils import ImporterDict

            scope.maps.append(ImporterDict())  # type: ignore

        # Note: Annotations can be strings inside of strings, like `"'int'"`. So evaluating it once
        # isn't necessarily enough; make sure to call `recurse()` with the result!
        if mode == "eval":
            try:
                # The globals must be a real dict, so the scope will be used as
                # the locals
                annotation = eval(annotation, {}, scope)
            except Exception:
                pass
            else:
                return recurse(annotation)
        elif mode == "getattr":
            first_name, *attrs = annotation.split(".")

            try:
                value = scope[first_name]
            except KeyError:
                pass
            else:
                try:
                    for attr in attrs:
                        value = getattr(value, attr)

                    return recurse(value)  # type: ignore
                except AttributeError:
                    pass
        elif mode == "ast":
            expr = ast.parse(annotation, mode="eval")
            try:
                result = _eval_ast(
                    expr.body,
                    scope,
                    strict=strict,
                    max_depth=max_depth,
                    treat_name_errors_as_imports=treat_name_errors_as_imports,
                )
            except Exception:
                pass
            else:
                return recurse(result)  # type: ignore
        else:
            assert False, f"Invalid mode: {mode!r}"

        if annotation == "ellipsis":
            return type(...)

        if not strict:
            return annotation

        raise CannotResolveForwardref(annotation, context)

    if annotation is None:
        return None

    # In some versions, `ParamSpec` is a subclass of `list`, so make sure this
    # check happens before the `isinstance(annotation, list)` check below
    if type(annotation) in (
        typing_extensions.TypeVarTuple,
        typing_extensions.ParamSpec,
        typing_extensions.ParamSpecKwargs,
        typing_extensions.ParamSpecArgs,
    ):
        return annotation

    if isinstance(annotation, t.TypeVar):
        if annotation.__constraints__:
            constraints = [recurse(constraint) for constraint in annotation.__constraints__]
            bound = None
        else:
            constraints = ()
            bound = recurse(annotation.__bound__)

        return t.TypeVar(
            annotation.__name__,  # type: ignore
            *constraints,  # type: ignore
            bound=bound,  # type: ignore
            covariant=annotation.__covariant__,  # type: ignore
            contravariant=annotation.__contravariant__,  # type: ignore
        )

    if not is_parameterized_generic(annotation, raising=False):
        return annotation

    base = get_generic_base_class(annotation)

    # Handle special cases where we can't blindly recurse into the subtypes
    if base in LITERAL_TYPES:
        return annotation

    if base is typing_extensions.Annotated:
        type_args = list(get_type_arguments(annotation))
        type_args[0] = recurse(type_args[0])  # type: ignore
        return parameterize(base, type_args)

    if base in (typing.Callable, collections.abc.Callable):
        arg_types, return_type = get_type_arguments(annotation)

        if isinstance(arg_types, list):
            arg_types = [recurse(typ) for typ in arg_types]

        return_type = recurse(return_type)  # type: ignore

        return parameterize(base, (arg_types, return_type))

    type_args = get_type_arguments(annotation)
    type_args = tuple(recurse(typ) for typ in type_args)  # type: ignore
    return parameterize(base, type_args)


def _eval_ast(
    node: ast.AST,
    scope: typing.Mapping[str, object],
    strict: bool,
    max_depth: int,
    treat_name_errors_as_imports: bool,
) -> object:
    # Compared to "eval" and "getattr", this method of evaluating forward refs
    # has the advantage of being able to perform partial evaluation. For
    # example, the forward ref `"ClassVar[NameThatCannotBeResolved]"` can be
    # turned into `ClassVar["NameThatCannotBeResolved"]`.

    def recurse(node: ast.AST) -> object:
        if max_depth <= 1:
            return ast.unparse(node)

        return _eval_ast(node, scope, strict, max_depth - 1, treat_name_errors_as_imports)

    if strict:
        safe_recurse = recurse
    else:

        def safe_recurse(node: ast.AST) -> object:
            try:
                return recurse(node)
            except Exception:
                return ast.unparse(node)

    def safe_recurse_if_forwardref(obj: object) -> object:
        if max_depth <= 1:
            return obj

        return resolve_forward_refs(
            obj,  # type: ignore
            scope,
            mode="ast",
            strict=strict,
            max_depth=max_depth - 1,
            treat_name_errors_as_imports=treat_name_errors_as_imports,
        )

    if type(node) is ast.Name:
        name = node.id
        return scope[name]
    elif type(node) is ast.Attribute:
        obj = recurse(node.value)

        try:
            return getattr(obj, node.attr)
        except AttributeError:
            # The parameter name is a little misleading, but we'll also treat AttributeErrors on
            # modules as missing imports
            if not treat_name_errors_as_imports or not isinstance(obj, types.ModuleType):
                raise

            try:
                importlib.import_module(f"{obj.__name__}.{node.attr}")
            except ImportError:
                pass

            try:
                return getattr(obj, node.attr)
            except AttributeError:
                pass

            raise
    elif type(node) is ast.Subscript:
        generic_type = recurse(node.value)
        subtype = safe_recurse(node.slice)
        return generic_type[subtype]  # type: ignore
    elif type(node) is ast.Constant:
        return node.value
    elif type(node) is ast.Tuple:
        return tuple(safe_recurse(elt) for elt in node.elts)
    elif type(node) is ast.List:
        return [safe_recurse(elt) for elt in node.elts]
    elif type(node) is ast.BinOp:
        if type(node.op) is ast.BitOr:
            left = safe_recurse(node.left)
            right = safe_recurse(node.right)
            # Use `Union` instead of `|` because:
            # 1. It works in older python versions
            # 2. It works even if `left` and `right` are strings because they
            #    couldn't be resolved
            return typing.Union[left, right]  # type: ignore

    raise NotImplementedError(f"Can't evaluate AST node {node}")


def annotation_to_string(
    annotation: TypeAnnotation,
    *,
    implicit_typing: bool = True,
    implicit_dataclasses: bool = True,
    new_style_unions: bool = True,
    optional_as_union: bool = True,
    variance_prefixes: bool = False,
    aliases: t.Mapping[object, str] = {},
) -> str:
    """
    Converts a type annotation to string. The result is valid python code
    (unless ``variance_prefixes`` is set to ``True``).

    Examples::

        >>> annotation_to_string(int)
        'int'
        >>> annotation_to_string(None)
        'None'
        >>> annotation_to_string(typing.List[int])
        'List[int]'

    :param annotation: A class or type annotation
    :param implicit_typing: Whether to omit the "typing." prefix from ``typing``
        types' names
    :param implicit_dataclasses: Whether to omit the "dataclasses." prefix from ``dataclasses.KW_ONLY``
    :param new_style_unions: Whether to use the new-style ``typing.Union`` syntax
        ``int | str`` instead of ``Union[int, str]``
    :param variance_prefixes: Whether `TypeVars` and `ParamSpecs` should be
        prefixed with ``+``, ``-``, or ``~`` to indicate variance
    :return: A string that, when evaluated, returns ``annotation``
    """

    def recurse(annotation: TypeAnnotation) -> str:
        return annotation_to_string(
            annotation,
            implicit_typing=implicit_typing,
            new_style_unions=new_style_unions,
            optional_as_union=optional_as_union,
            variance_prefixes=variance_prefixes,
            aliases=aliases,
        )

    def process_nested(prefix: str, elems: t.Iterable[TypeAnnotation]):
        elems = ", ".join(recurse(ann) for ann in elems)
        return "{}[{}]".format(prefix, elems)

    if isinstance(annotation, typing.ForwardRef):
        return _get_forward_ref_code(annotation)

    if isinstance(annotation, str):
        return annotation

    try:
        return aliases[annotation]
    except (KeyError, TypeError):
        pass

    if annotation is ...:
        return "..."

    if annotation is KW_ONLY:
        if implicit_dataclasses:
            return "KW_ONLY"
        else:
            return "dataclasses.KW_ONLY"

    if annotation in (None, type(None)):
        return "None"

    if is_parameterized_generic(annotation, raising=False):
        base = get_generic_base_class(annotation)
        subtypes = get_type_arguments(annotation)

        if base is typing.Optional and optional_as_union:
            base = typing.Union
            subtypes = [subtypes[0], None]

        if base is typing.Union and new_style_unions:
            return " | ".join(recurse(sub) for sub in subtypes)  # type: ignore

        if base in (typing.Callable, collections.abc.Callable):
            param_types, return_type = subtypes

            prefix = recurse(base)  # type: ignore[wtf]
            return_str = recurse(return_type)  # type: ignore

            if isinstance(param_types, list):
                params = ", ".join(recurse(param_type) for param_type in param_types)
                params = f"[{params}]"
            else:
                params = "..."

            return f"{prefix}[{params}, {return_str}]"

        if base in LITERAL_TYPES:
            literals = ", ".join(repr(value) for value in subtypes)
            prefix = recurse(base)  # type: ignore[wtf]
            return f"{prefix}[{literals}]"

        if base is typing_extensions.Annotated:
            sub_type, *annotations = subtypes
            sub_strs = [recurse(sub_type)]  # type: ignore
            sub_strs.extend(repr(ann) for ann in annotations)

            prefix = recurse(base)  # type: ignore[wtf]
            return f"{prefix}[{', '.join(sub_strs)}]"

        prefix = recurse(base)  # type: ignore[wtf]
        return process_nested(prefix, subtypes)  # type: ignore

    if isinstance(annotation, (typing.TypeVar, typing_extensions.ParamSpec)):
        result = annotation.__name__

        if variance_prefixes:
            if annotation.__covariant__:
                result = "+" + result
            elif annotation.__contravariant__:
                result = "-" + result
            else:
                result = "~" + result

        return result

    if isinstance(annotation, typing_extensions.TypeVarTuple):
        return annotation.__name__

    if isinstance(annotation, typing_extensions.ParamSpecArgs):
        return recurse(annotation.__origin__) + ".args"

    if isinstance(annotation, typing_extensions.ParamSpecKwargs):
        return recurse(annotation.__origin__) + ".kwargs"

    if hasattr(annotation, "__module__"):
        if annotation.__module__ == "builtins":
            return annotation.__qualname__
        elif annotation.__module__ in ("typing", "typing_extensions"):
            annotation = get_type_name(annotation)

            if not implicit_typing:
                annotation = "typing." + annotation

            return annotation
        else:
            return "{}.{}".format(annotation.__module__, annotation.__qualname__)

    return repr(annotation)  # This point should never be reached, but just in case...


def annotation_for_callable(callable_: typing.Callable[..., object]) -> Type_:
    """
    Given a callable object as input, returns a matching type annotation.

    Examples::

        >>> annotation_for_callable(len)
        typing.Callable[[typing.Sized], int]

    Note: How ``*args``, ``**kwargs``, and keyword-only parameters are handled
    is currently undefined.

    .. versionadded:: 1.5
    """
    signature = Signature.from_callable(callable_)
    parameters = list(signature.parameters.values())

    if signature.return_annotation is Signature.empty:
        return_type = typing.Any
    else:
        return_type = signature.return_annotation

    param_types = [
        typing.Any if param.annotation is Parameter.empty else param.annotation
        for param in parameters
        if param.kind <= Parameter.POSITIONAL_OR_KEYWORD
    ]  # TODO: Raise an exception if keyword-only parameters exist?

    # If some parameters are optional, we have to return a Union of Callable
    # types with fewer and fewer parameters
    options = [param_types]

    for param in reversed(parameters):
        if not param.is_optional:
            break

        param_types = param_types[:-1]
        options.append(param_types)

    if len(options) == 1:
        return typing.Callable[param_types, return_type]  # type: ignore

    options = tuple(typing.Callable[option, return_type] for option in options)  # type: ignore
    return typing.Union[options]  # type: ignore



import typing as t
import types

from introspection.typing import TypeInfo


def test_union():
    info = TypeInfo(int | str)
    assert info.type in (t.Union, types.UnionType)
    assert info.arguments == (int, str)



import typing
import typing_extensions as te


__all__ = [
    "DATACLASSES_KW_ONLY",
    "LITERAL_TYPES",
    "ANNOTATED_TYPES",
    "CLASS_VAR_TYPES",
    "TYPE_ALIAS_TYPES",
    "ANYS",
]


try:
    from dataclasses import KW_ONLY as DATACLASSES_KW_ONLY
except Exception:
    DATACLASSES_KW_ONLY = object()


LITERAL_TYPES = (te.Literal,)
if hasattr(typing, "Literal"):
    LITERAL_TYPES += (typing.Literal,)

ANNOTATED_TYPES = (te.Annotated,)
if hasattr(typing, "Annotated"):
    ANNOTATED_TYPES += (typing.Annotated,)

CLASS_VAR_TYPES = (te.ClassVar,)
if hasattr(typing, "ClassVar"):
    CLASS_VAR_TYPES += (typing.ClassVar,)

TYPE_ALIAS_TYPES = ()
if hasattr(typing, "TypeAliasType"):
    TYPE_ALIAS_TYPES += (typing.TypeAliasType,)  # type: ignore


# In python 3.10, `typing.Any` and `typing_extensions.Any` are two distinct objects.
ANYS = (typing.Any, te.Any)



import types
import dataclasses
from typing import *
import typing_extensions

from .types import *
from . import call_frame


__all__ = [
    "Error",
    "FunctionCallError",
    "ArgumentError",
    "ArgumentRequired",
    "InvalidArgumentType",
    "InvalidOption",
    "NotAClass",
    "NotAType",
    "NotAGeneric",
    "NotAParameterizedGeneric",
    "ForwardRefsDontHaveNames",
    "GenericMustNotBeParameterized",
    "ConflictingArguments",
    "SubTypeRequired",
    "NameNotAccessibleFromFrame",
    "DundermethodNotFound",
    "ObjectHasNoDict",
    "CannotUnwrapBoundMethod",
    "InvalidIdentifier",
    "NoSignatureFound",
    "MethodNotFound",
    "TypeVarNotSet",
    "NoConcreteTypeForTypeVar",
    "CannotResolveForwardref",
    "NoTypingEquivalent",
    "NoPythonEquivalent",
    "NoGenericPythonEquivalent",
]


T_co = TypeVar("T_co", covariant=True)
V_co = TypeVar("V_co", covariant=True)
M_co = TypeVar("M_co", covariant=True, bound=Union[type, Iterable[type]])


@typing_extensions.dataclass_transform()
class Error(Exception):
    def __init_subclass__(cls):
        dataclasses.dataclass(eq=False)(cls)


class FunctionCallError(Error):
    pass
    # function: Callable = dataclasses.field(init=False)

    # def __post_init__(self):
    #     self.function = 'TODO'


class ArgumentError(FunctionCallError):
    parameter: str


class InvalidArgumentError(Generic[V_co], ArgumentError):
    value: V_co

    def __str__(self) -> str:
        return f"Invalid value for parameter {self.parameter!r}: {self.value!r}"


class ArgumentRequired(ArgumentError):
    reason: str

    def __str__(self) -> str:
        return f"Because {self.reason}, the {self.parameter!r} argument is required"


class InvalidArgumentType(InvalidArgumentError[V_co], Generic[V_co, T_co], TypeError):
    expected_type: Type[T_co]

    def __str__(self) -> str:
        return super().__str__() + f". A value of type {self.expected_type!r} is required."


class InvalidOption(InvalidArgumentError[V_co], Generic[V_co, T_co], ValueError):
    options: Set[T_co]

    def __str__(self) -> str:
        return super().__str__() + f". Valid options are: {self.options}"


class NotAClass(InvalidArgumentError[V_co], TypeError):
    def __str__(self) -> str:
        return super().__str__() + ". A class is required."


class NotAType(InvalidArgumentError[V_co], TypeError):
    def __str__(self) -> str:
        return super().__str__() + ". A class or type is required."


class NotAGeneric(InvalidArgumentError[V_co], ValueError):
    def __str__(self) -> str:
        return super().__str__() + ". A generic type is required."


class NotAParameterizedGeneric(InvalidArgumentError[V_co], ValueError):
    def __str__(self) -> str:
        return super().__str__() + ". A parameterized generic is required."


class ForwardRefsDontHaveNames(InvalidArgumentError[V_co], TypeError):
    def __str__(self) -> str:
        return super().__str__() + ". ForwardRefs don't have names."


class GenericMustNotBeParameterized(InvalidArgumentError[V_co], ValueError):
    def __str__(self) -> str:
        return super().__str__() + ". Only unparameterized generics are allowed."


class ObjectHasNoDict(InvalidArgumentError[V_co], TypeError):
    def __str__(self) -> str:
        return super().__str__() + ". It doesn't have a __dict__."


class ConflictingArguments(FunctionCallError, TypeError):
    parameter1: str
    parameter2: str

    def __str__(self) -> str:
        return f"You cannot pass {self.parameter1!r} and {self.parameter2!r} at the same time"


class SubTypeRequired(FunctionCallError, ValueError):
    type: Type_
    base_type: Type_

    def __str__(self) -> str:
        return f"{self.type!r} isn't a subtype of {self.base_type!r}"


class NameNotAccessibleFromFrame(Error, NameError):
    name: str
    frame: "call_frame.CallFrame"

    def __str__(self) -> str:
        return f"Name {self.name!r} is not accessible from frame {self.frame!r}"


class DundermethodNotFound(Error, AttributeError):
    method_name: str
    cls: type

    def __str__(self) -> str:
        return f"Class {self.cls!r} doesn't implement {self.method_name}"


class CannotUnwrapBoundMethod(Error, TypeError):
    method: types.MethodType

    def __str__(self) -> str:
        return "Cannot unwrap a bound method"


class InvalidIdentifier(Error, NameError):
    identifier: str

    def __str__(self) -> str:
        return f"Failed to resolve identifier {self.identifier!r}"


class NoSignatureFound(Error, ValueError):
    callable: Callable[..., object]

    def __str__(self) -> str:
        return f"Can't determine signature of {self.callable!r}"


class MethodNotFound(Error, AttributeError, Generic[M_co]):
    method_name: str
    class_or_mro: M_co

    def __str__(self) -> str:
        return f"No method named {self.method_name!r} found in {self.class_or_mro}"


class TypeVarNotSet(Error):
    type_var: TypeVar
    base_type: type
    type: type

    def __str__(self) -> str:
        return f"TypeVar {self.type_var} of {self.base_type!r} is not set in {self.type!r}"


class NoConcreteTypeForTypeVar(Error):
    type_var: TypeVar
    base_type: type
    type: type
    final_var: TypeVar

    def __str__(self) -> str:
        return (
            f"TypeVar {self.type_var} of {self.base_type!r} doesn't have a concrete value in"
            f" {self.type!r}; it is set to {self.final_var!r}"
        )


class CannotResolveForwardref(Error, ValueError):
    forward_ref: ForwardReference
    context: Optional[ForwardRefContext]

    def __str__(self) -> str:
        return f"Cannot resolve forward reference {self.forward_ref!r} in context {self.context!r}"


class NoTypingEquivalent(Error, ValueError):
    type: type

    def __str__(self) -> str:
        return f"{self.type!r} has no typing equivalent"


class NoPythonEquivalent(Error, ValueError):
    type: Type_

    def __str__(self) -> str:
        return f"{self.type!r} has no plain python equivalent"


class NoGenericPythonEquivalent(Error, ValueError):
    type: Type_

    def __str__(self) -> str:
        return f"{self.type!r} has no (generic) plain python equivalent"



import pytest

import sys
import typing as t

from introspection.typing import is_instance, is_subtype


T = t.TypeVar("T")


class AwaitableObject:
    def __await__(self) -> t.Iterator[None]: ...


def func_with_forwardrefs(arg: "int") -> "str": ...


@pytest.mark.parametrize(
    "obj, type_, expected",
    [
        (3, int, True),
        (True, int, True),
        (3, float, True),
        (True, float, True),
        (b"", float, False),
        (1.5, int, False),
        ("hi", str, True),
        ("hi", t.AnyStr, True),
        ("hi", T, True),
        (b"", str, False),
        ([], tuple, False),
        ([], list, True),
        ([], t.List[int], True),
        ([1], t.List[int], True),
        ([True], t.List[float], True),
        ([1], t.List[str], False),
        (1, t.Union[int, str], True),
        ("hi", t.Union[int, str], True),
        (b"", t.Union[int, str], False),
        (None, t.Optional[int], True),
        (1, t.Optional[int], True),
        (b"", t.Optional[int], False),
        (1, t.Annotated[int, "hi"], True),
        ((), tuple, True),
        ((), t.Tuple, True),
        ((), t.Tuple[int], False),
        ((1, "str"), t.Tuple[int, str], True),
        ((), t.Tuple[int, ...], True),
        ((1, 2), t.Tuple[int, ...], True),
        ((1, b""), t.Tuple[int, ...], False),
        (dict, t.Callable[[], t.Any], True),
        # (list, Callable[[str], list], True),
        # (list, Callable[[str], List[str]], True),
        (AwaitableObject(), t.Awaitable, True),
        (AwaitableObject(), t.Awaitable[t.Any], True),
        (AwaitableObject(), t.Awaitable[object], True),
        # (AwaitableObject(), Awaitable[int], False),
        (func_with_forwardrefs, t.Callable[[int], str], True),
        (func_with_forwardrefs, t.Callable[[float], str], False),
        (func_with_forwardrefs, t.Callable[[int], bytes], False),
    ],
)
def test_is_instance(obj, type_, expected):
    assert is_instance(obj, type_) == expected


# @pytest.mark.parametrize(
#     "obj, type_, expected",
#     [
#         (3, "ThisIsAnInvalidForwardRef", False),
#         # (3, "datetime", False),  # This is interesting because `datetime` is a module
#     ],
# )
# def test_is_instance_with_forwardref_type(obj, type_, expected):
#     assert is_instance(obj, type_, treat_name_errors_as_imports=True) == expected


@pytest.mark.parametrize(
    "subtype, supertype, expected",
    [
        (dict, t.Any, True),
        (t.Any, dict, True),
        (dict, t.Callable, False),
        (tuple, t.Iterable, True),
        # (List[bool], Sequence[int], True),
        (tuple, t.Union[list, tuple], True),
        (tuple, t.Union[list, t.Iterable], True),
        (t.Iterable, t.Union[list, str], False),
        (tuple, t.Optional[tuple], True),
        (tuple, t.Optional[list], False),
    ],
)
def test_is_subtype(subtype, supertype, expected):
    assert is_subtype(subtype, supertype) == expected


if sys.version_info >= (3, 10):

    @pytest.mark.parametrize(
        "subtype, supertype, expected",
        [
            (tuple, list | tuple, True),
            (tuple, list | t.Iterable, True),
            (t.Iterable, list | str, False),
            (tuple, tuple | None, True),
            (tuple, list | None, False),
        ],
    )
    def test_is_subtype_py30(subtype, supertype, expected):
        assert is_subtype(subtype, supertype) == expected



import types
import typing
import typing_extensions

import sentinel


__all__ = [
    "Case",
    "TypeParameter",
    "Type_",
    "ParameterizedGeneric",
    "ForwardReference",
    "TypeAnnotation",
    "ForwardRefContext",
]


NONE = sentinel.create("NONE")


_GenericAliases = [
    (types, "GenericAlias"),
    (typing, "_GenericAlias"),
]
GenericAliases = tuple(
    getattr(module, name) for module, name in _GenericAliases if hasattr(module, name)
)


T = typing.TypeVar("T")
P = typing_extensions.ParamSpec("P")
Class = typing.TypeVar("Class", bound=type)

Case = typing.Literal["snake", "upper snake", "camel", "pascal", "kebab", "upper kebab"]
TypeParameter = typing.Union[typing.TypeVar, typing_extensions.TypeVarTuple]
Function = types.FunctionType
ParameterizedGeneric: typing_extensions.TypeAlias = "types.GenericAlias"
DataclassesKwOnly = object  # The stdlib couldn't even bother making `KW_ONLY` a class. Sigh.
Type_ = typing.Union[type, typing.TypeVar, ParameterizedGeneric, None, DataclassesKwOnly]
ForwardReference = typing.Union[str, typing.ForwardRef]
TypeAnnotation = typing.Union[Type_, ForwardReference]
ForwardRefContext = typing.Union[
    typing.Callable, types.ModuleType, str, typing.Mapping[str, object]
]


# class _MetaGeneric(typing_extensions.Protocol[T]):  # type: ignore
#     def __getitem__(self, subtypes: typing.Union[T, typing.Tuple[T, ...]]) -> ParameterizedGeneric:
#         ...


# class _ClassGeneric(typing_extensions.Protocol[T]):  # type: ignore
#     def __class_getitem__(
#         cls, subtypes: typing.Union[T, typing.Tuple[T, ...]]
#     ) -> ParameterizedGeneric:
#         ...


# GenericType = typing.Union[_MetaGeneric[T], typing.Type[_ClassGeneric[T]]]


class Slot(typing.Protocol[T]):  # type: ignore[variance]
    def __get__(self, instance: T, owner: typing.Optional[typing.Type[T]]) -> object: ...

    def __set__(self, instance: T, value: object) -> None: ...

    def __delete__(self, instance: T) -> None: ...


class ObjectWithQualname(typing.Protocol):
    __name__: str
    __qualname__: str



import dataclasses
import sys
import typing as t

from .introspection import _get_forward_ref_code

if sys.version_info >= (3, 14):
    import annotationlib
else:
    annotationlib = None

from ._compat import CLASS_VAR_TYPES, DATACLASSES_KW_ONLY
from .forward_ref import ForwardRef
from .type_info import TypeInfo
from ..errors import CannotResolveForwardref
from ..misc import static_mro, static_vars
from ..types import TypeAnnotation


__all__ = ["get_type_annotations", "get_instance_attribute_annotations"]


@t.overload
def get_type_annotations(
    cls: type, *, allow_forwardrefs: t.Literal[False] = False
) -> t.Dict[str, TypeInfo]: ...


@t.overload
def get_type_annotations(
    cls: type, *, allow_forwardrefs: bool
) -> t.Dict[str, t.Union[TypeInfo, ForwardRef]]: ...


def get_type_annotations(  # type: ignore (wtf? Some variance nonsense?)
    cls: type, *, allow_forwardrefs: bool = False
) -> t.Dict[str, t.Union[TypeInfo, ForwardRef]]:
    """
    Similar to `typing.get_type_hints`, but returns the annotations as `TypeInfo` objects.
    """
    result: t.Dict[str, t.Union[TypeInfo, ForwardRef]] = {}

    for class_ in static_mro(cls):
        cls_vars = static_vars(class_)
        try:
            annotations = _get_local_annotations_dict(cls_vars)
        except KeyError:
            continue

        for attr_name, annotation in annotations.items():
            if attr_name in result:
                continue

            try:
                value = TypeInfo(annotation, forward_ref_context=class_)
            except CannotResolveForwardref:
                if not allow_forwardrefs:
                    raise

                value = ForwardRef(annotation, class_)  # type: ignore

            result[attr_name] = value

    return result


def _get_local_annotations_dict(
    cls_dict: t.Mapping[str, t.Any],
) -> t.Dict[str, TypeAnnotation]:
    if annotationlib is not None:
        annotate = annotationlib.get_annotate_from_class_namespace(cls_dict)
        if annotate is not None:
            return annotationlib.call_annotate_function(annotate, annotationlib.Format.FORWARDREF)

    return cls_dict.get("__annotations__", {})


@t.overload
def get_instance_attribute_annotations(
    cls: type, *, allow_forwardrefs: t.Literal[False] = False
) -> t.Dict[str, TypeInfo]: ...


@t.overload
def get_instance_attribute_annotations(
    cls: type, *, allow_forwardrefs: bool
) -> t.Dict[str, t.Union[TypeInfo, ForwardRef]]: ...


def get_instance_attribute_annotations(  # type: ignore (wtf? Some variance nonsense?)
    cls: type, *, allow_forwardrefs: bool = False
) -> t.Dict[str, t.Union[TypeInfo, ForwardRef]]:
    """
    Similar to `get_type_annotations`, but it automatically removes `_: dataclasses.KW_ONLY`
    and `typing.ClassVar` annotations.

    .. versionadded:: 1.13
    """
    result: t.Dict[str, t.Union[TypeInfo, ForwardRef]] = {}

    for name, info in get_type_annotations(cls, allow_forwardrefs=allow_forwardrefs).items():
        if isinstance(info, TypeInfo):
            if info.type is DATACLASSES_KW_ONLY or info.type in CLASS_VAR_TYPES:
                continue
        else:
            forward_ref_code = _get_forward_ref_code(info)
            if forward_ref_code in (
                "KW_ONLY",
                "dataclasses.KW_ONLY",
            ) or forward_ref_code.startswith(
                (
                    "ClassVar",
                    "typing.ClassVar",
                    "typing_extensions.ClassVar",
                    "t.ClassVar",
                    "te.ClassVar",
                )
            ):
                continue

        result[name] = info

    return result
